@@ -40,7 +40,12 @@ import kernel_tuner.core as core
 from kernel_tuner.runners.sequential import SequentialRunner
 from kernel_tuner.runners.simulation import SimulationRunner
 
-from kernel_tuner.strategies import brute_force, random_sample, diff_evo, minimize, basinhopping, genetic_algorithm, pso, simulated_annealing, firefly_algorithm, bayes_opt
+try:
+    import torch
+except ImportError:
+    torch = util.TorchPlaceHolder()
+
+from kernel_tuner.strategies import brute_force, random_sample, diff_evo, minimize, basinhopping, genetic_algorithm, mls, pso, simulated_annealing, firefly_algorithm, bayes_opt
 
 strategy_map = {
     "brute_force": brute_force,
@@ -49,10 +54,11 @@ strategy_map = {
     "basinhopping": basinhopping,
     "diff_evo": diff_evo,
     "genetic_algorithm": genetic_algorithm,
+    "mls": mls,
     "pso": pso,
     "simulated_annealing": simulated_annealing,
     "firefly_algorithm": firefly_algorithm,
-    "bayes_opt": bayes_opt
+    "bayes_opt": bayes_opt,
 }
 
 
@@ -404,6 +410,7 @@ def tune_kernel(kernel_name, kernel_string, problem_size, arguments, tune_params
     kernel_options = Options([(k, opts[k]) for k in _kernel_options.keys()])
     tuning_options = Options([(k, opts[k]) for k in _tuning_options.keys()])
     device_options = Options([(k, opts[k]) for k in _device_options.keys()])
+    tuning_options["snap"] = True
 
     logging.debug('tune_kernel called')
     logging.debug('kernel_options: %s', util.get_config_string(kernel_options))
@@ -441,8 +448,8 @@ def tune_kernel(kernel_name, kernel_string, problem_size, arguments, tune_params
         strategy = brute_force
 
     # select the runner for this job based on input
-    SelectedRunner = SimulationRunner if simulation_mode else SequentialRunner
-    with SelectedRunner(kernel_source, kernel_options, device_options, iterations, observers) as runner:
+    selected_runner = SimulationRunner if simulation_mode is True else SequentialRunner
+    with selected_runner(kernel_source, kernel_options, device_options, iterations, observers) as runner:
 
         #the user-specified function may or may not have an optional atol argument;
         #we normalize it so that it always accepts atol.
@@ -544,6 +551,9 @@ def run_kernel(kernel_name, kernel_string, problem_size, arguments, params, grid
             if func is None:
                 raise Exception("cannot compile kernel, too much shared memory used")
 
+            #add shared memory arguments to compiled module
+            if smem_args is not None:
+                dev.copy_shared_memory_args(util.get_smem_args(smem_args, params))
             #add constant memory arguments to compiled module
             if cmem_args is not None:
                 dev.copy_constant_memory_args(cmem_args)
@@ -564,6 +574,8 @@ def run_kernel(kernel_name, kernel_string, problem_size, arguments, params, grid
         for i, arg in enumerate(arguments):
             if numpy.isscalar(arg):
                 results.append(arg)
+            elif isinstance(arg, torch.Tensor):
+                results.append(arg.cpu())
             else:
                 results.append(numpy.zeros_like(arg))
                 dev.memcpy_dtoh(results[-1], gpu_args[i])
